@@ -1,4 +1,4 @@
-/* Copyright (c) 1998-2005 Thorsten Kukuk
+/* Copyright (c) 1998-2009 Thorsten Kukuk
    This file is part of ypbind-mt.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
@@ -50,6 +50,8 @@
 #include "local.h"
 #include "pthread_np.h"
 
+extern int verbose_flag;
+
 #if (defined(__sun__) || defined(sun)) && defined(__svr4__)
 typedef uint32_t u_int32_t;
 #endif
@@ -78,8 +80,25 @@ struct binding
   bool_t use_broadcast;
   struct bound_server server[_MAXSERVER];
   struct bound_server ypset;
-  CLIENT *client_handle;
+  struct sockaddr_in sin;
+  struct bound_server last; /* last written */
 };
+static inline char *
+bound_host(struct binding *bptr)
+{
+  struct bound_server *sptr;
+
+  if (bptr->active >= 0)
+  	sptr = &bptr->server[bptr->active];
+  else if (bptr->active == -2)
+  	sptr = &bptr->ypset;
+  else
+  	return "Unknown Host";
+
+  if (sptr->host != NULL)
+  	return(sptr->host);
+  return (inet_ntoa(sptr->addr));
+}
 
 static struct binding *domainlist = NULL;
 static int max_domains = 0;
@@ -90,13 +109,13 @@ static void do_broadcast (struct binding *list);
 static int ping_all (struct binding *list);
 
 static void
-remove_bindingfile (const char *domain)
+remove_bindingfile (const char *domain_name)
 {
-  char path[strlen (BINDINGDIR) + strlen (domain) + 10];
+  char path[strlen (BINDINGDIR) + strlen (domain_name) + 10];
 
-  sprintf (path, "%s/%s.1", BINDINGDIR, domain);
+  sprintf (path, "%s/%s.1", BINDINGDIR, domain_name);
   unlink (path);
-  sprintf (path, "%s/%s.2", BINDINGDIR, domain);
+  sprintf (path, "%s/%s.2", BINDINGDIR, domain_name);
   unlink (path);
 }
 
@@ -115,26 +134,44 @@ update_bindingfile (struct binding *entry)
   sprintf (path1, "%s/%s.1", BINDINGDIR, entry->domain);
   sprintf (path2, "%s/%s.2", BINDINGDIR, entry->domain);
 
-  iov[0].iov_base = (caddr_t) &sport;
-  iov[0].iov_len = sizeof (sport);
-  iov[1].iov_base = (caddr_t) &ybres;
-  iov[1].iov_len = sizeof ybres;
-
   memset(&ybres, 0, sizeof (ybres));
   ybres.ypbind_status = YPBIND_SUCC_VAL;
   if (entry->active >= 0)
     {
+      if (entry->last.host &&
+          !memcmp(&entry->server[entry->active].addr, &entry->last.addr,
+		  sizeof(struct in_addr)) &&
+          entry->server[entry->active].port == entry->last.port)
+        {
+	  if (debug_flag)
+	    log_msg (LOG_DEBUG, "Entry for %s unchanged, skipping writeout",
+		     entry->domain);
+          return;
+        }
+
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_addr,
 	      &entry->server[entry->active].addr, sizeof (struct in_addr));
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_port,
 	      &entry->server[entry->active].port, sizeof (unsigned short int));
+      entry->last= entry->server[entry->active];
     }
   else if (entry->active == -2) /* ypset was used */
     {
+      if (entry->last.host &&
+          !memcmp(&entry->ypset.addr, &entry->last.addr,
+		  sizeof(struct in_addr)) &&
+          entry->ypset.port == entry->last.port)
+        {
+	  if (debug_flag)
+	    log_msg (LOG_DEBUG, "Entry for %s unchanged, skipping writeout",
+		     entry->domain);
+          return;
+        }
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_addr,
 	      &entry->ypset.addr, sizeof (struct in_addr));
       memcpy (&ybres.ypbind_resp_u.ypbind_bindinfo.ypbind_binding_port,
 	      &entry->ypset.port, sizeof (unsigned short int));
+      entry->last= entry->ypset;
     }
   else
     {
@@ -142,10 +179,16 @@ update_bindingfile (struct binding *entry)
           libc will query ypbind direct. */
        unlink (path1);
        unlink (path2);
+       entry->last.host = NULL;
        log_msg (LOG_ERR, "INTERNAL ERROR: update_bindingfile called without valid data!");
        return;
     }
 
+  iov[0].iov_base = (caddr_t) &sport;
+  iov[0].iov_len = sizeof (sport);
+  iov[1].iov_base = (caddr_t) &ybres;
+  iov[1].iov_len = sizeof ybres;
+
   len = iov[0].iov_len + iov[1].iov_len;
 
   if ((fd = open(path1, O_CREAT | O_RDWR | O_TRUNC, FILE_MODE )) != -1)
@@ -153,6 +196,7 @@ update_bindingfile (struct binding *entry)
       if (writev (fd, iov, 2) != len )
         {
           log_msg (LOG_ERR, "writev (%s): %s", path1, strerror (errno));
+	  entry->last.host = NULL;
           unlink (path1);
         }
       close (fd);
@@ -165,6 +209,7 @@ update_bindingfile (struct binding *entry)
       if (writev (fd, iov, 2) != len )
         {
           log_msg (LOG_ERR, "writev (%s): %s", path2, strerror (errno));
+	   entry->last.host = NULL;
           unlink (path2);
         }
       close (fd);
@@ -186,15 +231,10 @@ change_binding (const char *domain, ypbind_binding *binding)
       if (strcmp (domainlist[i].domain, domain) == 0)
 	{
 	  struct sockaddr_in addr;
-	  struct timeval timeout;
-	  int sock;
 
 	  pthread_rdwr_runlock_np (&domainlock);
 	  pthread_rdwr_wlock_np (&domainlock);
 
-	  if (domainlist[i].client_handle != NULL)
-	    clnt_destroy (domainlist[i].client_handle);
-
 	  domainlist[i].active = -2;
 	  memcpy(&(domainlist[i].ypset.addr),
 		 &(binding->ypbind_binding_addr),
@@ -204,26 +244,29 @@ change_binding (const char *domain, ypbind_binding *binding)
 		 sizeof (unsigned short int));
 	  domainlist[i].ypset.family = AF_INET;
 
-	  sock = RPC_ANYSOCK;
-	  timeout.tv_sec = 1;
-	  timeout.tv_usec = 0;
-	  memset (&addr, 0, sizeof (struct sockaddr_in));
-	  memcpy (&addr.sin_addr, &domainlist[i].ypset.addr,
-		  sizeof (struct in_addr));
-	  memcpy (&addr.sin_port, &domainlist[i].ypset.port,
-		  sizeof (unsigned short int));
-	  addr.sin_family = domainlist[i].ypset.family;
-
-	  if ((domainlist[i].client_handle =
-	       clntudp_create(&addr, YPPROG, YPVERS, timeout, &sock)) == NULL)
-	    {
-	      domainlist[i].active = -1;
-	      remove_bindingfile (domain);
-	    }
+          memset (&addr, 0, sizeof (struct sockaddr_in));
+          memcpy (&addr.sin_addr, &domainlist[i].ypset.addr,
+                  sizeof (struct in_addr));
+          memcpy (&addr.sin_port, &domainlist[i].ypset.port,
+                  sizeof (unsigned short int));
+          addr.sin_family = domainlist[i].ypset.family;
+	  memcpy (&domainlist[i].sin, &addr, sizeof (struct sockaddr_in));
 	  pthread_rdwr_wunlock_np (&domainlock);
 	  pthread_rdwr_rlock_np (&domainlock);
 	  update_bindingfile (&domainlist[i]);
 	  pthread_rdwr_runlock_np (&domainlock);
+	  if (verbose_flag)
+	    {
+	      log_msg (LOG_NOTICE, "NIS server set to '%s'"
+		" for domain '%s'",
+	        bound_host(&domainlist[i]), domainlist[i].domain);
+	    }
+	  if (logfile_flag && (logfile_flag & LOG_SERVER_CHANGES))
+	    {
+	      log2file ("NIS server for domain '%s' set to '%s' ",
+			domainlist[i].domain,
+			bound_host(&domainlist[i]));
+	    }
 
 	  return;
 	}
@@ -330,15 +373,18 @@ clear_server (void)
     {
       for (i = 0; i < max_domains; ++i)
 	{
-	  for (j = 0; j < _MAXSERVER; ++j)
+	  if (domainlist[i].active != -1)
 	    {
-	      if (domainlist[i].server[j].host != NULL)
-		free (domainlist[i].server[j].host);
+	      remove_bindingfile (domainlist[i].domain);
+	      for (j = 0; j < _MAXSERVER; ++j)
+		{
+		  if (domainlist[i].server[j].host != NULL)
+		    free (domainlist[i].server[j].host);
+		}
+	      if (domainlist[i].ypset.host != NULL)
+		free (domainlist[i].ypset.host);
+	      domainlist[i].active = -1;
 	    }
-	  if (domainlist[i].ypset.host != NULL)
-	    free (domainlist[i].ypset.host);
-	  if (domainlist[i].client_handle != NULL)
-	    clnt_destroy (domainlist[i].client_handle);
 	}
       free (domainlist);
     }
@@ -374,6 +420,7 @@ get_entry (const char *domain, struct binding **entry)
       domainlist[max_domains - 1].ypset.host = NULL;
       domainlist[max_domains - 1].active = (-1);
       domainlist[max_domains - 1].use_broadcast = FALSE;
+      domainlist[max_domains - 1].last.host = NULL;
       memset (domainlist[max_domains - 1].server, 0,
 	      (_MAXSERVER * sizeof (struct bound_server)));
       *entry = &domainlist[max_domains - 1];
@@ -427,7 +474,7 @@ add_server (const char *domain, const char *host, int check_syntax)
 	  break;
 
       /* There is no empty slot */
-      if (entry->server[active].host != NULL)
+      if (active >= _MAXSERVER)
 	{
 	  if (debug_flag)
 	    log_msg (LOG_DEBUG,
@@ -524,9 +571,6 @@ static struct binding *in_use = NULL;
 static bool_t
 eachresult (bool_t *out, struct sockaddr_in *addr)
 {
-  struct timeval timeout;
-  int sock;
-
   if (*out)
     {
       if(debug_flag)
@@ -589,15 +633,7 @@ eachresult (bool_t *out, struct sockaddr_in *addr)
 	     sizeof (struct in_addr));
       memcpy(&(in_use->server[0].port), &addr->sin_port,
 	     sizeof (unsigned short int));
-
-      sock = RPC_ANYSOCK;
-      timeout.tv_sec = 1;
-      timeout.tv_usec = 0;
-      in_use->client_handle =
-	clntudp_create(addr, YPPROG, YPVERS, timeout, &sock);
-
-      if (in_use->client_handle == NULL)
-	return 0;
+      memcpy (&(in_use->sin), &addr, sizeof (struct sockaddr_in));
 
       in_use->active = 0;
 
@@ -756,6 +792,7 @@ ping_all (struct binding *list)
   u_long i, pings_count = 0;
   struct cu_data *cu;
   char *domain = list->domain;
+  int old_active = list->active;
 
   if (list->server[0].host == NULL) /* There is no known server */
     return 0;
@@ -792,6 +829,10 @@ ping_all (struct binding *list)
       list->server[i].port = s_in.sin_port;
       if (s_in.sin_port == 0)
 	{
+	  if (verbose_flag && list->active == i)
+		log_msg (LOG_NOTICE, "NIS server '%s' not repsonding "
+		    "for domain '%s'", list->server[i].host, list->domain);
+
 	  if (debug_flag)
 	    log_msg (LOG_DEBUG, _("host '%s' doesn't answer."),
 		     list->server[i].host);
@@ -848,36 +889,53 @@ ping_all (struct binding *list)
   clnt_call (clnt, YPPROC_DOMAIN_NONACK, (xdrproc_t) NULL, (caddr_t) NULL,
              (xdrproc_t) xdr_bool, (caddr_t) &clnt_res, TIMEOUT00);
 
-  xid_lookup = *((u_int32_t *) (cu->cu_inbuf));
+  memcpy (&xid_lookup, &(cu->cu_inbuf), sizeof (u_int32_t));
   close (sock);
   for (i = 0; i < pings_count; ++i)
     {
       if (pings[i]->xid == xid_lookup)
         {
+	  CLIENT *client_handle;
 	  pthread_rdwr_wlock_np (&domainlock);
 
 	  sock = RPC_ANYSOCK;
-	  list->client_handle =
-	    clntudp_create (&(pings[i]->sin),
-			    YPPROG, YPVERS, TIMEOUT50, &sock);
-	  if (list->client_handle == NULL)
+	  client_handle = clntudp_create (&(pings[i]->sin),
+					  YPPROG, YPVERS, TIMEOUT50, &sock);
+	  if (client_handle == NULL)
 	    {
 	      /* NULL should not happen, we have got an answer from the server. */
 	      log_msg (LOG_DEBUG,
-		       _("Server '%s' for domain '%s' answer to ping but failed to bind"),
+		       _("Server '%s' for domain '%s' answered ping but failed to bind"),
 		       list->server[list->active].host, domain);
 	    }
 	  else
 	    {
+	      clnt_destroy (client_handle);
+
+	      memcpy (&(list->sin), &pings[i]->sin, sizeof(struct sockaddr_in));
 	      list->active = pings[i]->server_nr;
 	      pthread_rdwr_wunlock_np (&domainlock);
 	      pthread_rdwr_rlock_np (&domainlock);
 	      update_bindingfile (list);
 	      pthread_rdwr_runlock_np (&domainlock);
+
 	      if (debug_flag)
 		log_msg (LOG_DEBUG,
 			 _("Answer for domain '%s' from server '%s'"),
 			 domain, list->server[list->active].host);
+
+	      if (logfile_flag && (logfile_flag & LOG_SERVER_CHANGES) &&
+		  old_active != list->active)
+		{
+		  if (old_active == -1)
+		    log2file ("NIS server for domain '%s' set to '%s'",
+			      domain, list->server[list->active].host);
+		  else
+		    log2file ("NIS server for domain '%s' changed from '%s' to '%s'",
+			      domain, list->server[old_active].host,
+			      list->server[list->active].host);
+		}
+
 	      found = 1;
 	    }
         }
@@ -909,6 +967,7 @@ ping_all (struct binding *list)
   struct timeval timeout;
   CLIENT *clnt_handlep = NULL;
   int i = 0;
+  int old_active = list->active;
 
   if (list->server[0].host == NULL) /* There is no known server */
     return 0;
@@ -971,13 +1030,31 @@ ping_all (struct binding *list)
         {
 	  memcpy (&(list->server[i].port), &server_addr.sin_port,
 		  sizeof (unsigned short int));
-          list->client_handle = clnt_handlep;
+	  memcpy (&(list->sin), &server_addr, sizeof (struct sockaddr_in));
+          clnt_destroy (clnt_handlep);
           pthread_rdwr_wlock_np (&domainlock);
           list->active = i;
           pthread_rdwr_wunlock_np (&domainlock);
           pthread_rdwr_rlock_np (&domainlock);
           update_bindingfile (list);
           pthread_rdwr_runlock_np (&domainlock);
+	  if (debug_flag)
+	    log_msg (LOG_DEBUG,
+		     _("Answer for domain '%s' from server '%s'"),
+		     domain, list->server[list->active].host);
+
+	  if (logfile_flag && (logfile_flag & LOG_SERVER_CHANGES) &&
+	      old_active != list->active)
+	    {
+	      if (old_active == -1)
+		log2file ("NIS server for domain '%s' set to '%s'",
+			  domain, list->server[list->active].host);
+	      else
+		log2file ("NIS server for domain '%s' changed from '%s' to '%s'",
+			  domain, list->server[old_active].host,
+			  list->server[list->active].host);
+	    }
+
           return 1;
         }
 
@@ -996,13 +1073,21 @@ ping_all (struct binding *list)
 void
 do_binding (void)
 {
-  int i;
+  int i, active;
 
   pthread_mutex_lock (&search_lock);
   for (i = 0; i < max_domains; ++i)
     {
+      active = domainlist[i].active;
+
       if (!ping_all (&domainlist[i]) && domainlist[i].use_broadcast)
 	do_broadcast (&domainlist[i]);
+      if (verbose_flag &&
+		domainlist[i].active >= 0 && active != domainlist[i].active)
+	{
+		log_msg (LOG_NOTICE, "NIS server is '%s' for domain '%s'",
+		    bound_host(&domainlist[i]), domainlist[i].domain);
+    	}
     }
   pthread_mutex_unlock (&search_lock);
 }
@@ -1016,7 +1101,10 @@ test_bindings (void *param __attribute__ ((unused)))
   static int success = 0;
   int lastcheck = 0;
 
-  do_binding ();
+#ifdef USE_DBUS_NM
+  if (is_online)
+#endif
+    do_binding ();
 
   if (ping_interval < 1)
     pthread_exit (&success);
@@ -1025,8 +1113,7 @@ test_bindings (void *param __attribute__ ((unused)))
     {
       sleep (ping_interval);
 
-      /* Check, if ping_interval was changed through a SIGHUP.
-	 Not possible in the moment, but maybe in the future. */
+      /* Check, if ping_interval was changed through a SIGHUP.  */
       if (ping_interval < 1)
 	pthread_exit (&success);
 
@@ -1034,15 +1121,18 @@ test_bindings (void *param __attribute__ ((unused)))
       if (lastcheck >= 900) /* 900 = 15min. */
 	lastcheck = 0;
 
-      lastcheck = test_bindings_once (lastcheck, NULL);
+#if USE_DBUS_NM
+      if (is_online)
 
+#endif
+	lastcheck = test_bindings_once (lastcheck, NULL);
     } /* end while() endless loop */
 }
 
 int
 test_bindings_once (int lastcheck, const char *req_domain)
 {
-  int i;
+  int i, active;
 
   /* Since we need the write lock later, getting the read lock here is
      not enough. During the time, where we wait for the write lock, the
@@ -1052,9 +1142,9 @@ test_bindings_once (int lastcheck, const char *req_domain)
   if (debug_flag)
     {
       if (lastcheck)
-	log_msg (LOG_DEBUG, _("Pinging all active server."));
+	log_msg (LOG_DEBUG, _("Pinging all active servers."));
       else
-	log_msg (LOG_DEBUG, _("Check new for fastest server."));
+	log_msg (LOG_DEBUG, _("Checking for new fastest server."));
     }
 
   for (i = 0; i < max_domains; ++i)
@@ -1071,13 +1161,7 @@ test_bindings_once (int lastcheck, const char *req_domain)
 	  continue;
 	}
 
-      /* We should never run into this. For debugging.  */
-      if (domainlist[i].client_handle == NULL && domainlist[i].active != -1)
-	{
-	  log_msg (LOG_ERR, "ALERT: active=%d, but client_handle is NULL!",
-		   domainlist[i].active);
-	  domainlist[i].active = -1;
-	}
+      active = domainlist[i].active;
 
       if (domainlist[i].active != -1)
 	{
@@ -1085,18 +1169,41 @@ test_bindings_once (int lastcheck, const char *req_domain)
 	     the fastest one. */
 	  if (lastcheck != 0)
 	    {
-	      /* Check only if the current binding is still valid. */
-	      struct timeval time_out;
+	      const struct timeval TIMEOUT50 = {5, 0};
+	      int sock = RPC_ANYSOCK;
+	      CLIENT *client_handle =
+		clntudp_create (&(domainlist[i].sin),
+				YPPROG, YPVERS, TIMEOUT50, &sock);
+	      if (client_handle == NULL)
+		{
+		  if (verbose_flag)
+		    log_msg (LOG_NOTICE,
+			     "NIS server '%s' for domain '%s' not reachable",
+			     bound_host(&domainlist[i]),
+			     domainlist[i].domain);
+		  status = RPC_CANTSEND;
+		}
+	      else
+		{
+		  /* Check only if the current binding is still valid. */
+		  struct timeval time_out;
 
-	      time_out.tv_sec = 3;
-	      time_out.tv_usec = 0;
-	      status =
-		clnt_call(domainlist[i].client_handle,
-			  YPPROC_DOMAIN, (xdrproc_t) ypbind_xdr_domainname,
-			  (caddr_t) &domain, (xdrproc_t) xdr_bool,
-			  (caddr_t) &out, time_out);
+		  time_out.tv_sec = 3;
+		  time_out.tv_usec = 0;
+		  status =
+		    clnt_call(client_handle, YPPROC_DOMAIN,
+			      (xdrproc_t) ypbind_xdr_domainname,
+			      (caddr_t) &domain,
+			      (xdrproc_t) xdr_bool,
+			      (caddr_t) &out, time_out);
+		  if (verbose_flag && status != RPC_SUCCESS)
+		    log_msg (LOG_NOTICE,
+			     "NIS server '%s' not responding for domain '%s'",
+			     bound_host(&domainlist[i]),
+			     domainlist[i].domain);
+		  clnt_destroy (client_handle);
+		}
 	    }
-
 	  /* time to search a new fastest server, but only if the current
 	     one was not set with ypset. We search in every case if the
 	     above check fails and the current data is not longer valid. */
@@ -1105,39 +1212,7 @@ test_bindings_once (int lastcheck, const char *req_domain)
 	    {
 	      /* The current binding is not valid or it is time to search
 		 for a new, fast server. */
-	      if (debug_flag && lastcheck != 0)
-		{
-		  /* Current active binding is not longer valid, print
-		     the old binding for debugging. */
-		  if (domainlist[i].use_broadcast)
-		    log_msg (LOG_DEBUG,
-			     _("Server for domain '%s' doesn't answer."),
-			     domain);
-		  else
-		    {
-		      if (domainlist[i].active == -2)
-			log_msg (LOG_DEBUG,
-				 _("Server '%s' for domain '%s' doesn't answer."),
-				 inet_ntoa(domainlist[i].ypset.addr),
-				 domain);
-		      else
-			log_msg (LOG_DEBUG,
-				 _("Server '%s' for domain '%s' doesn't answer."),
-				 domainlist[i].server[domainlist[i].active].host,
-				 domain);
-		    }
-		}
-	      /* We can destroy the client_handle since we are the
-		 only thread who uses it. */
-	      /* client_handle can be NULL? */
-	      if (domainlist[i].client_handle == NULL)
-		{
-		  log_msg (LOG_ERR, "ALERT: client_handle=NULL, active=%d, lastcheck=%d, domain=%s",
-			   domainlist[i].active, lastcheck, domain);
-		}
-	      else
-		clnt_destroy (domainlist[i].client_handle);
-	      domainlist[i].client_handle = NULL;
+
 	      if (domainlist[i].active == -2)
 		{
 		  /* We can give this free, server does not answer any
@@ -1146,31 +1221,39 @@ test_bindings_once (int lastcheck, const char *req_domain)
 		    free (domainlist[i].ypset.host);
 		  domainlist[i].ypset.host = NULL;
 		}
-	      domainlist[i].active = -1;
 	      lastcheck = 0; /* If we need a new server before the TTL expires,
 				reset it. */
 	      /* And give the write lock away, search a new host and get
 		 the write lock again. */
 	      pthread_rdwr_wunlock_np (&domainlock);
 	      pthread_mutex_lock (&search_lock);
-	      if (!ping_all (&domainlist[i]) &&
-		  domainlist[i].use_broadcast)
+	      if (domainlist[i].use_broadcast)
 		do_broadcast (&domainlist[i]);
+	      else
+		ping_all (&domainlist[i]);
 	      pthread_mutex_unlock (&search_lock);
 	      pthread_rdwr_wlock_np (&domainlock);
 	    }
 	}
-      else
+      else /* domainlist[i].active == -1 */
 	{
 	  /* there is no binding for this domain, try to find a new
 	     server */
 	  pthread_rdwr_wunlock_np (&domainlock);
 	  pthread_mutex_lock (&search_lock);
-	  if (!ping_all (&domainlist[i]) && domainlist[i].use_broadcast)
+	  if (domainlist[i].use_broadcast)
 	    do_broadcast (&domainlist[i]);
+	  else
+	    ping_all (&domainlist[i]);
 	  pthread_mutex_unlock (&search_lock);
 	  pthread_rdwr_wlock_np (&domainlock);
 	}
+      if (verbose_flag &&
+          domainlist[i].active >= 0 && active != domainlist[i].active)
+	{
+	  log_msg (LOG_NOTICE, "NIS server is '%s' for domain '%s'",
+	      bound_host(&domainlist[i]), domainlist[i].domain);
+	}
     } /* end for () all domains */
 
   pthread_rdwr_wunlock_np (&domainlock);
