@@ -1,4 +1,4 @@
-/* Copyright (c) 1998, 1999, 2001, 2002, 2004, 2005, 2006 Thorsten Kukuk
+/* Copyright (c) 1998 - 2009 Thorsten Kukuk
    This file is part of ypbind-mt.
    Author: Thorsten Kukuk <kukuk@suse.de>
 
@@ -100,7 +100,7 @@ unlink_bindingdir (void)
     }
 }
 
-/* Load the config file (/etc/yp.conf)  */
+/* Load or check syntax of the config file (/etc/yp.conf)  */
 int
 load_config (int check_syntax)
 {
@@ -243,8 +243,10 @@ load_config (int check_syntax)
 		++have_entries;
 	      else
 		++bad_entries;
-	      continue;
 	    }
+	  else
+	    ++bad_entries;
+          continue;
 	}
       else if (strncmp (cp, "broadcast", 9) == 0 &&
 	       (isspace ((int)cp[9]) || cp[9] == '\0'))
@@ -297,6 +299,25 @@ load_config (int check_syntax)
   return 0;
 }
 
+
+/* Load the configuration, exiting if there's an error */
+static void
+load_config_or_exit(void)
+{
+  if (load_config (0) != 0)
+    {
+      fputs (_("No NIS server and no -broadcast option specified.\n"),
+	     stderr);
+      fprintf (stderr,
+	       _("Add a NIS server to the %s configuration file,\n"),
+	       configfile);
+      fputs (_("or start ypbind with the -broadcast option.\n"),
+	     stderr);
+      exit (1);
+    }
+}
+
+
 /* Create a pidfile on startup */
 static void
 create_pidfile (void)
@@ -440,6 +461,9 @@ sig_handler (void *v_param  __attribute__ ((unused)))
 	  close (lock_fd);
 	  unlink (_YPBIND_PIDFILE);
 	  unlink_bindingdir ();
+	  if (logfile_flag)
+	    log2file ("Stopping %s-%s", PACKAGE, VERSION);
+	  close_logfile ();
 	  exit (0);
 	  break;
 	case SIGHUP:
@@ -466,6 +490,7 @@ sig_handler (void *v_param  __attribute__ ((unused)))
 	}
     }
 }
+int verbose_flag;
 
 static void
 usage (int ret)
@@ -478,7 +503,7 @@ usage (int ret)
     output = stdout;
 
   fputs (_("Usage:\n"), output);
-  fputs (_("\typbind [-broadcast | -ypset | -ypsetme] [-p port] [-f configfile]\n\t  [-no-ping] [-broken-server] [-local-only] [-i ping-interval] [-debug]\n"), output);
+  fputs (_("\typbind [-broadcast | -ypset | -ypsetme] [-f configfile]\n\t  [-no-ping] [-broken-server] [-local-only] [-i ping-interval] [-debug] [-verbose]\n"), output);
 #ifdef USE_DBUS_NM
   fputs (_("\t  [-no-dbus]\n"), output);
 #endif
@@ -557,6 +582,7 @@ portmapper_register (void)
     {
       log_msg (LOG_ERR,
 	       _("Unable to register (YPBINDPROG, YPBINDVERS, udp)."));
+      svc_destroy (transp);
       return 1;
     }
 
@@ -565,6 +591,7 @@ portmapper_register (void)
     {
       log_msg (LOG_ERR,
 	       _("Unable to register (YPBINDPROG, YPBINDOLDVERS, udp)."));
+      svc_destroy (transp);
       return 1;
     }
 
@@ -574,6 +601,7 @@ portmapper_register (void)
       if (sock < 0)
 	{
 	  log_msg (LOG_ERR, _("Cannot create TCP: %s"), strerror (errno));
+	  svc_destroy (transp);
 	  return 1;
 	}
 
@@ -592,6 +620,8 @@ portmapper_register (void)
       if (result < 0)
 	{
 	  log_msg (LOG_ERR, _("Cannot bind TCP: %s"), strerror (errno));
+	  close (sock);
+	  svc_destroy (transp);
 	  return 1;
 	}
     }
@@ -610,6 +640,7 @@ portmapper_register (void)
 		     IPPROTO_TCP))
     {
       log_msg (LOG_ERR, _("Unable to register (YPBINDPROG, YPBINDVERS, tcp)."));
+      svc_destroy (transp);
       return 1;
     }
 
@@ -618,6 +649,7 @@ portmapper_register (void)
     {
       log_msg (LOG_ERR,
 	       _("Unable to register (YPBINDPROG, YPBINDOLDVERS, tcp)."));
+      svc_destroy (transp);
       return 1;
     }
   return 0;
@@ -656,6 +688,9 @@ main (int argc, char **argv)
       else if (strcmp ("-d", argv[i]) == 0 ||
 	       strcmp ("-debug", argv[i]) == 0)
         debug_flag = 1;
+      else if (strcmp ("-v", argv[i]) == 0 ||
+	       strcmp ("-verbose", argv[i]) == 0)
+        verbose_flag = 1;
       else if (strcmp ("-broken-server", argv[i]) == 0 ||
 	       strcmp ("-broken_server", argv[i]) == 0)
         broken_server = 1;
@@ -682,7 +717,7 @@ main (int argc, char **argv)
 	  port = atoi (argv[i]);
 	}
       else if (strcmp ("-ping-interval", argv[i]) == 0 ||
-	       strcmp ("-ping-interval", argv[i]) == 0 ||
+	       strcmp ("-ping_interval", argv[i]) == 0 ||
 	       strcmp ("-i", argv[i]) == 0)
 	{
 	  if (i+1 == argc || argv[i+1][0] == '-')
@@ -692,6 +727,13 @@ main (int argc, char **argv)
 	}
       else if (strcmp ("-c", argv[i]) == 0)
 	configcheck_only = 1;
+      else if (strcmp ("-log", argv[i]) == 0)
+	{
+	  if (i+1 == argc || argv[i+1][0] == '-')
+	    usage (1);
+	  ++i;
+	  logfile_flag = atoi (argv[i]);
+	}
 #ifdef USE_DBUS_NM
       else if (strcmp ("-no-dbus", argv[i]) == 0)
 	disable_dbus = 1;
@@ -764,15 +806,17 @@ main (int argc, char **argv)
 
   if (!use_broadcast)
     {
-      if (load_config (0) != 0)
-	{
-	  fputs (_("No NIS server and no -broadcast option specified.\n"), stderr);
-	  fprintf (stderr,
-		   _("Add a NIS server to the %s configuration file,\n"),
-		   configfile);
-	  fputs (_("or start ypbind with the -broadcast option.\n"), stderr);
-	  exit (1);
-	}
+#ifdef USE_DBUS_NM
+      /* If we don't use DBUS, exit with an error if we cannot load the
+	 config. Else load the config, maybe there is a network already
+	 running. */
+      if (disable_dbus)
+	load_config_or_exit ();
+      else
+	load_config (0);
+#else
+      load_config_or_exit ();
+#endif
     }
   else
     add_server (domain, NULL, 0);
@@ -820,8 +864,16 @@ main (int argc, char **argv)
 
       umask (0);
       j = open ("/dev/null", O_RDWR);
-      dup (j);
-      dup (j);
+      if (dup (j) == -1)
+	{
+          log_msg (LOG_ERR, "Cannot dup file handle: %s\n", strerror (errno));
+          exit (-1);
+	}
+      if (dup (j) == -1)
+	{
+          log_msg (LOG_ERR, "Cannot dup file handle: %s\n", strerror (errno));
+          exit (-1);
+	}
     }
 
 #if defined(HAVE___NSS_CONFIGURE_LOOKUP)
@@ -869,6 +921,9 @@ main (int argc, char **argv)
     is_online = 1;
 #endif
 
+  if (logfile_flag)
+    log2file ("Starting %s-%s", PACKAGE, VERSION);
+
   portmapper_disconnect ();
   if (portmapper_register () != 0)
     {
